from abc import abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from functools import wraps
from importlib.util import find_spec

from typing import Any, Mapping, ParamSpec, TypeVar, Callable, Protocol, List


P = ParamSpec('P')
V = TypeVar('V')


__ALL__ = ['requires']


class DecoratedCallable(Protocol):
    __globals__: Mapping[str, Any]

    @abstractmethod
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> V:
        pass


@dataclass
class ModuleInfo:
    name: str
    package: str | None
    pypi_name: str | None

    def __str__(self):
        return self.pypi_name or self.name


ModuleArg = str | ModuleInfo | Sequence[str] | Sequence[str, str | None] | Sequence[str, str | None, str | None]


def _parse_module_arg(arg: ModuleArg | Sequence[ModuleArg], package: str = None,
                      pypi_name: str = None) -> ModuleInfo:

    match arg:
        case module if isinstance(module, ModuleInfo):
            return module
        case module_name if isinstance(module_name, str):
            return ModuleInfo(module_name, package or None, pypi_name or None)
        case [module_name, package]:
            return ModuleInfo(module_name, package or None, None)
        case [module_name, package, pypi_name]:
            return ModuleInfo(module_name, package or None, pypi_name or None)
        case _:
            raise TypeError(f"{arg} is not a valid module argument: "
                            f"(module_name, package | None, pypi_name | None)")


def _parse_args(modules: Sequence[ModuleArg], package: str = None, pypi_name: str = None) -> List[ModuleInfo]:
    if pypi_name and len(modules) > 1:
        raise ValueError("pypi_name can't be specified as a *kwarg when modules is a sequence.")
    return [_parse_module_arg(module, package, pypi_name) for module in modules]


def _message_not_found(modules: List[ModuleInfo], caller_name) -> str:
    if len(modules) == 1:
        return f"{modules[0]} is not installed and is required by {caller_name}"
    else:
        modules_txt = ", ".join(str(m) for m in modules)
        return f"{caller_name} requires modules that are not installed: {modules_txt}"


def requires(*modules: ModuleArg, package: str = None, pypi_name: str = None, runtime_check=True
             ) -> Callable[[DecoratedCallable], DecoratedCallable]:

    modules = _parse_args(modules, package, pypi_name)

    def decorate(f: DecoratedCallable) -> DecoratedCallable:
        check = not all(m for m in modules if not find_spec(m.name, package=m.package))

        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> V:
            nonlocal check

            if not check:
                not_found = [m for m in modules if not f.__globals__.get(m.name, None)]
                if not_found:
                    error_msg = _message_not_found(not_found, f.__name__)
                    raise ModuleNotFoundError(error_msg)
            else:
                check = True

            return f(*args, **kwargs)

        return wrapper

    return decorate



from importlib.util import find_spec
from io import StringIO
from pathlib import Path
from typing import Any

from .decorators import requires


if find_spec('charset_normalizer'):
    import charset_normalizer
else:
    charset_normalizer = None


__ALL__ = ["detect_file_encoding", "decode_file", "detect_bytes_encoding", "decode_bytes",
           "EncodingDetectionException", ]


class EncodingDetectionException(Exception):
    pass


def _decode_with(decoder, obj, **decoding_kwargs) -> charset_normalizer.CharsetMatches:
    results = decoder(obj, **decoding_kwargs)
    if not results:
        raise EncodingDetectionException(f"{decoder} unable to detect encoding from {obj!r}")
    return results


@requires('charset_normalizer')
def detect_file_encoding(fp: Path | str, **decoding_kwargs: Any) -> str:
    """
    Detects the file type and returns a file-like object with the contents of the file.
    :param fp: path to the file or file-like object.
    :return: file-like object with the contents of the file.
    """
    results = _decode_with(charset_normalizer.from_path, Path(fp), **decoding_kwargs)
    return results.best().encoding


@requires('charset_normalizer')
def decode_file(fp: Path | str, **decoding_kwargs: Any) -> StringIO:
    """
    Detects the file type and returns a file-like object with the contents of the file.
    :param fp: path to the file or file-like object.
    :return: file-like object with the contents of the file.
    """
    results = _decode_with(charset_normalizer.from_path, Path(fp), **decoding_kwargs)
    return StringIO(str(results.best()))


@requires('charset_normalizer')
def decode_bytes(b: bytes | bytearray, **decoding_kwargs: Any) -> str:
    results = charset_normalizer.from_bytes(b, **decoding_kwargs)
    return str(results.best())


@requires('charset_normalizer')
def detect_bytes_encoding(b: bytes | bytearray, **decoding_kwargs: Any) -> str:
    results = charset_normalizer.from_bytes(b, **decoding_kwargs)
    return results.best().encoding



# Code extracted from nv.utils in order to reduce dependency footprint for nv.tables

import sys
from collections.abc import MutableSet, Set, Sequence
from collections import OrderedDict
from itertools import repeat
from typing import  Any, Collection, List, TypeVar, Iterable, Iterator, Mapping, Dict


__ALL__ = ["extract_headers", "normalize", "check_normalized", "normalize_row", ]


_KEY_ONLY = object()

V = TypeVar('V')


class OrderedSet(MutableSet[V]):
    """
    Since CPython 3.6, dictionaries are ordered at implementation and tend to be much faster than any pure
    Python implementation. For versions older than 3.6, OrderedDict is used instead.

    OrderedSet is a wrapper over an ordered dict. Set values are stored as keys. All values point to the same
    dummy object (_KEY_ONLY).
    """

    OrderedMapping = dict if sys.version_info >= (3, 6) else OrderedDict

    def __init__(self, items: Iterable[V] | None = None):
        self.__mapping = self.OrderedMapping(zip(items, repeat(_KEY_ONLY))) if items else self.OrderedMapping()

    def add(self, value: V):
        self.__mapping.update({value: _KEY_ONLY})

    def discard(self, value: V):
        del self.__mapping[value]

    def __contains__(self, x: object) -> bool:
        return x in self.__mapping

    def __len__(self) -> int:
        return len(self.__mapping)

    def __iter__(self) -> Iterator[V]:
        return iter(self.__mapping)

    def __eq__(self, other):
        if isinstance(other, OrderedSet | Sequence):
            return len(self) == len(other) and list(self) == list(other)
        elif isinstance(other, Set):
            return set(self) == other
        else:
            return NotImplemented

    def __repr__(self):
        return f'{self.__class__.__name__}({list(self)!r})'


def check_normalized(data: Iterable[Mapping[str, Any]], headers: Collection[str] = None) -> bool:
    """
    Checks if a list of dictionaries is normalized.
    :param data: Iterable with dictionaries or dataclasses
    :param headers: Collection of strings with headers. If None, headers are extracted from the first row.
    :return: True if all rows have the same keys, False otherwise.
    """
    headers = set(headers) if headers else None
    for row in data:
        this_headers = set(row.keys())

        # First header
        if headers is None:
            headers = this_headers
            continue

        if this_headers != headers:
            return False

    return True


def extract_headers(data: Iterable[Mapping[str, Any]], preserve_order: bool = True) -> List[str]:
    """
    Extracts all keys from a list of dictionaries. If preserve_order is True, the algorithm tries to preserve the
    natural order of keys (as dictionaries are ordered since Python 3.6+ and column headers tends to have some
    organization logic behind its order). Set it to false for a quicker execution that does not attempt to preserve
    natural order.
    :param data: data as an iterable of key-value dictionaries.
    :param preserve_order:
    :return:
    """
    headers = None
    natural_order = False

    for row in data:
        this_headers = OrderedSet(row.keys()) if natural_order else set(row.keys())

        # First header
        if headers is None:
            headers = this_headers
            natural_order = True
            continue

        # Straight forward implementation of headers
        if not preserve_order:
            headers = headers | this_headers
            continue

        # Order preservation heuristics
        extra_items = this_headers - headers
        missing_items = headers - this_headers

        if not missing_items and (extra_items or not natural_order):
            # Either a superset of current headers or a natural equal was found
            headers = this_headers
            natural_order = True
            continue

        if extra_items:
            # Stack up extra items and continue to look for a better natural proxy
            headers = headers | extra_items
            natural_order = False

    return list(headers)


def normalize_row(row: Mapping[str, Any], headers: Collection[str], default: Any = None) -> Dict[str, Any]:
    """
    Normalizes a single dictionary, so that all items in the list have same keys by setting non-existent fields to
    default.
    :param row: dictionary.
    :param headers: collection of strings with headers (order will be preserved if possible).
    :param default: default value to be set to non-existent items (typically either None or empty string)
    :return: normalized copy of dictionary.
    """
    return {k: row.get(k, default) for k in headers}


def iter_normalize(data: Iterable[Mapping[str, Any]], headers: Collection[str] | None = None, default: Any = None,
                   ) -> Iterator[Dict[str, Any]]:
    for row in data:
        yield normalize_row(row, headers, default)


def normalize(data: Iterable[Mapping[str, Any]], headers: Collection[str] = None,
              default: Any = None) -> List[Dict[str, Any]]:
    """
    Normalizes a list of dictionaries, so that all items in the list have same keys by setting non-existent fields to
    default.
    :param data: list of dictionaries.
    :param headers: collection of strings with headers (order will be preserved if possible).
    :param default: default value to be set to non existent items (typically either None or empty string)
    :return: normalized copy of list.
    """
    if not headers:
        raise ValueError("Headers are required to normalize data")

    return list(iter_normalize(data, headers, default))



import dataclasses
from typing import Dict, Any, Mapping, List, Iterable, Tuple, ClassVar, Protocol


class DataClass(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, 'dataclasses.Field']]
    __dataclass_params__: ClassVar[Any]


# Generic data structures
DataRow = Mapping[str, Any] | DataClass
DataTable = Iterable[DataRow]
DataStructure = Iterable[Tuple[str, DataTable]] | Mapping[str, DataTable]


# Output data structures
ParsedDataRow = Dict[str, Any] | DataClass
ParsedDataTable = List[ParsedDataRow]
ParsedDataStructure = Mapping[str, ParsedDataTable]



from abc import abstractmethod
from functools import wraps
from importlib.util import find_spec
from pathlib import Path
from typing import Collection, Any, Mapping, Type, Tuple, ParamSpec, TypeVar, Callable, Iterator, Protocol

from .typing import DataTable, DataStructure, ParsedDataRow, ParsedDataTable, ParsedDataStructure


class Serializer:
    @abstractmethod
    def write_table(self, table: DataTable, fp: Path | str, headers: Collection[str] = None, default: Any = None,
                    **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def append_table(self, table: DataTable, fp: Path | str, headers: Collection[str] = None, default: Any = None,
                     **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def dump(self, structure: DataStructure, fp: Path | str, schema: Mapping[str, Collection[str]], default: Any = None,
             **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def iter_table(self, fp: Path | str, constructor: Type = None, default: Any = None,
                   **parser_kwargs) -> Iterator[ParsedDataRow]:
        raise NotImplementedError

    @abstractmethod
    def iter_structure(self, fp: Path | str, constructor: Type = None, default: Any = None,
                       **parser_kwargs) -> Iterator[Tuple[str, ParsedDataRow]]:
        raise NotImplementedError

    @abstractmethod
    def read_table(self, fp: Path | str, constructor: Type = None, default: Any = None,
                   **parser_kwargs) -> ParsedDataTable:
        raise NotImplementedError

    @abstractmethod
    def load(self, fp: Path | str, constructor: Type = None, default: Any = None,
             **parser_kwargs) -> ParsedDataStructure:
        raise NotImplementedError

    @abstractmethod
    def read_headers(self, fp: Path | str, **parser_kwargs) -> Collection[str]:
        pass

    @abstractmethod
    def read_schema(self, fp: Path | str, **parser_kwargs) -> Mapping[str, Collection[str]]:
        pass
